# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Common DNSSEC-related functions and constants."""

from io import BytesIO
import struct
import time

import dns.exception
import dns.name
import dns.node
import dns.rdataset
import dns.rdata
import dns.rdatatype
import dns.rdataclass
from ._compat import string_types


class UnsupportedAlgorithm(dns.exception.DNSException):
    """The DNSSEC algorithm is not supported."""


class ValidationFailure(dns.exception.DNSException):
    """The DNSSEC signature is invalid."""


#: RSAMD5
RSAMD5 = 1
#: DH
DH = 2
#: DSA
DSA = 3
#: ECC
ECC = 4
#: RSASHA1
RSASHA1 = 5
#: DSANSEC3SHA1
DSANSEC3SHA1 = 6
#: RSASHA1NSEC3SHA1
RSASHA1NSEC3SHA1 = 7
#: RSASHA256
RSASHA256 = 8
#: RSASHA512
RSASHA512 = 10
#: ECDSAP256SHA256
ECDSAP256SHA256 = 13
#: ECDSAP384SHA384
ECDSAP384SHA384 = 14
#: INDIRECT
INDIRECT = 252
#: PRIVATEDNS
PRIVATEDNS = 253
#: PRIVATEOID
PRIVATEOID = 254

_algorithm_by_text = {
    'RSAMD5': RSAMD5,
    'DH': DH,
    'DSA': DSA,
    'ECC': ECC,
    'RSASHA1': RSASHA1,
    'DSANSEC3SHA1': DSANSEC3SHA1,
    'RSASHA1NSEC3SHA1': RSASHA1NSEC3SHA1,
    'RSASHA256': RSASHA256,
    'RSASHA512': RSASHA512,
    'INDIRECT': INDIRECT,
    'ECDSAP256SHA256': ECDSAP256SHA256,
    'ECDSAP384SHA384': ECDSAP384SHA384,
    'PRIVATEDNS': PRIVATEDNS,
    'PRIVATEOID': PRIVATEOID,
}

# We construct the inverse mapping programmatically to ensure that we
# cannot make any mistakes (e.g. omissions, cut-and-paste errors) that
# would cause the mapping not to be true inverse.

_algorithm_by_value = {y: x for x, y in _algorithm_by_text.items()}


def algorithm_from_text(text):
    """Convert text into a DNSSEC algorithm value.

    Returns an ``int``.
    """

    value = _algorithm_by_text.get(text.upper())
    if value is None:
        value = int(text)
    return value


def algorithm_to_text(value):
    """Convert a DNSSEC algorithm value to text

    Returns a ``str``.
    """

    text = _algorithm_by_value.get(value)
    if text is None:
        text = str(value)
    return text


def _to_rdata(record, origin):
    s = BytesIO()
    record.to_wire(s, origin=origin)
    return s.getvalue()


def key_id(key, origin=None):
    """Return the key id (a 16-bit number) for the specified key.

    Note the *origin* parameter of this function is historical and
    is not needed.

    Returns an ``int`` between 0 and 65535.
    """

    rdata = _to_rdata(key, origin)
    rdata = bytearray(rdata)
    if key.algorithm == RSAMD5:
        return (rdata[-3] << 8) + rdata[-2]
    else:
        total = 0
        for i in range(len(rdata) // 2):
            total += (rdata[2 * i] << 8) + \
                rdata[2 * i + 1]
        if len(rdata) % 2 != 0:
            total += rdata[len(rdata) - 1] << 8
        total += ((total >> 16) & 0xffff)
        return total & 0xffff


def make_ds(name, key, algorithm, origin=None):
    """Create a DS record for a DNSSEC key.

    *name* is the owner name of the DS record.

    *key* is a ``dns.rdtypes.ANY.DNSKEY``.

    *algorithm* is a string describing which hash algorithm to use.  The
    currently supported hashes are "SHA1" and "SHA256".  Case does not
    matter for these strings.

    *origin* is a ``dns.name.Name`` and will be used as the origin
    if *key* is a relative name.

    Returns a ``dns.rdtypes.ANY.DS``.
    """

    if algorithm.upper() == 'SHA1':
        dsalg = 1
        hash = SHA1.new()
    elif algorithm.upper() == 'SHA256':
        dsalg = 2
        hash = SHA256.new()
    else:
        raise UnsupportedAlgorithm('unsupported algorithm "%s"' % algorithm)

    if isinstance(name, string_types):
        name = dns.name.from_text(name, origin)
    hash.update(name.canonicalize().to_wire())
    hash.update(_to_rdata(key, origin))
    digest = hash.digest()

    dsrdata = struct.pack("!HBB", key_id(key), key.algorithm, dsalg) + digest
    return dns.rdata.from_wire(dns.rdataclass.IN, dns.rdatatype.DS, dsrdata, 0,
                               len(dsrdata))


def _find_candidate_keys(keys, rrsig):
    candidate_keys = []
    value = keys.get(rrsig.signer)
    if value is None:
        return None
    if isinstance(value, dns.node.Node):
        try:
            rdataset = value.find_rdataset(dns.rdataclass.IN,
                                           dns.rdatatype.DNSKEY)
        except KeyError:
            return None
    else:
        rdataset = value
    for rdata in rdataset:
        if rdata.algorithm == rrsig.algorithm and \
                key_id(rdata) == rrsig.key_tag:
            candidate_keys.append(rdata)
    return candidate_keys


def _is_rsa(algorithm):
    return algorithm in (RSAMD5, RSASHA1,
                         RSASHA1NSEC3SHA1, RSASHA256,
                         RSASHA512)


def _is_dsa(algorithm):
    return algorithm in (DSA, DSANSEC3SHA1)


def _is_ecdsa(algorithm):
    return _have_ecdsa and (algorithm in (ECDSAP256SHA256, ECDSAP384SHA384))


def _is_md5(algorithm):
    return algorithm == RSAMD5


def _is_sha1(algorithm):
    return algorithm in (DSA, RSASHA1,
                         DSANSEC3SHA1, RSASHA1NSEC3SHA1)


def _is_sha256(algorithm):
    return algorithm in (RSASHA256, ECDSAP256SHA256)


def _is_sha384(algorithm):
    return algorithm == ECDSAP384SHA384


def _is_sha512(algorithm):
    return algorithm == RSASHA512


def _make_hash(algorithm):
    if _is_md5(algorithm):
        return MD5.new()
    if _is_sha1(algorithm):
        return SHA1.new()
    if _is_sha256(algorithm):
        return SHA256.new()
    if _is_sha384(algorithm):
        return SHA384.new()
    if _is_sha512(algorithm):
        return SHA512.new()
    raise ValidationFailure('unknown hash for algorithm %u' % algorithm)


def _make_algorithm_id(algorithm):
    if _is_md5(algorithm):
        oid = [0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x02, 0x05]
    elif _is_sha1(algorithm):
        oid = [0x2b, 0x0e, 0x03, 0x02, 0x1a]
    elif _is_sha256(algorithm):
        oid = [0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01]
    elif _is_sha512(algorithm):
        oid = [0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03]
    else:
        raise ValidationFailure('unknown algorithm %u' % algorithm)
    olen = len(oid)
    dlen = _make_hash(algorithm).digest_size
    idbytes = [0x30] + [8 + olen + dlen] + \
              [0x30, olen + 4] + [0x06, olen] + oid + \
              [0x05, 0x00] + [0x04, dlen]
    return struct.pack('!%dB' % len(idbytes), *idbytes)


def _validate_rrsig(rrset, rrsig, keys, origin=None, now=None):
    """Validate an RRset against a single signature rdata

    The owner name of *rrsig* is assumed to be the same as the owner name
    of *rrset*.

    *rrset* is the RRset to validate.  It can be a ``dns.rrset.RRset`` or
    a ``(dns.name.Name, dns.rdataset.Rdataset)`` tuple.

    *rrsig* is a ``dns.rdata.Rdata``, the signature to validate.

    *keys* is the key dictionary, used to find the DNSKEY associated with
    a given name.  The dictionary is keyed by a ``dns.name.Name``, and has
    ``dns.node.Node`` or ``dns.rdataset.Rdataset`` values.

    *origin* is a ``dns.name.Name``, the origin to use for relative names.

    *now* is an ``int``, the time to use when validating the signatures,
    in seconds since the UNIX epoch.  The default is the current time.
    """

    if isinstance(origin, string_types):
        origin = dns.name.from_text(origin, dns.name.root)

    candidate_keys = _find_candidate_keys(keys, rrsig)
    if candidate_keys is None:
        raise ValidationFailure('unknown key')

    for candidate_key in candidate_keys:
        # For convenience, allow the rrset to be specified as a (name,
        # rdataset) tuple as well as a proper rrset
        if isinstance(rrset, tuple):
            rrname = rrset[0]
            rdataset = rrset[1]
        else:
            rrname = rrset.name
            rdataset = rrset

        if now is None:
            now = time.time()
        if rrsig.expiration < now:
            raise ValidationFailure('expired')
        if rrsig.inception > now:
            raise ValidationFailure('not yet valid')

        hash = _make_hash(rrsig.algorithm)

        if _is_rsa(rrsig.algorithm):
            keyptr = candidate_key.key
            (bytes_,) = struct.unpack('!B', keyptr[0:1])
            keyptr = keyptr[1:]
            if bytes_ == 0:
                (bytes_,) = struct.unpack('!H', keyptr[0:2])
                keyptr = keyptr[2:]
            rsa_e = keyptr[0:bytes_]
            rsa_n = keyptr[bytes_:]
            try:
                pubkey = CryptoRSA.construct(
                    (number.bytes_to_long(rsa_n),
                     number.bytes_to_long(rsa_e)))
            except ValueError:
                raise ValidationFailure('invalid public key')
            sig = rrsig.signature
        elif _is_dsa(rrsig.algorithm):
            keyptr = candidate_key.key
            (t,) = struct.unpack('!B', keyptr[0:1])
            keyptr = keyptr[1:]
            octets = 64 + t * 8
            dsa_q = keyptr[0:20]
            keyptr = keyptr[20:]
            dsa_p = keyptr[0:octets]
            keyptr = keyptr[octets:]
            dsa_g = keyptr[0:octets]
            keyptr = keyptr[octets:]
            dsa_y = keyptr[0:octets]
            pubkey = CryptoDSA.construct(
                (number.bytes_to_long(dsa_y),
                 number.bytes_to_long(dsa_g),
                 number.bytes_to_long(dsa_p),
                 number.bytes_to_long(dsa_q)))
            sig = rrsig.signature[1:]
        elif _is_ecdsa(rrsig.algorithm):
            # use ecdsa for NIST-384p -- not currently supported by pycryptodome

            keyptr = candidate_key.key

            if rrsig.algorithm == ECDSAP256SHA256:
                curve = ecdsa.curves.NIST256p
                key_len = 32
            elif rrsig.algorithm == ECDSAP384SHA384:
                curve = ecdsa.curves.NIST384p
                key_len = 48

            x = number.bytes_to_long(keyptr[0:key_len])
            y = number.bytes_to_long(keyptr[key_len:key_len * 2])
            if not ecdsa.ecdsa.point_is_valid(curve.generator, x, y):
                raise ValidationFailure('invalid ECDSA key')
            point = ecdsa.ellipticcurve.Point(curve.curve, x, y, curve.order)
            verifying_key = ecdsa.keys.VerifyingKey.from_public_point(point,
                                                                      curve)
            pubkey = ECKeyWrapper(verifying_key, key_len)
            r = rrsig.signature[:key_len]
            s = rrsig.signature[key_len:]
            sig = ecdsa.ecdsa.Signature(number.bytes_to_long(r),
                                        number.bytes_to_long(s))

        else:
            raise ValidationFailure('unknown algorithm %u' % rrsig.algorithm)

        hash.update(_to_rdata(rrsig, origin)[:18])
        hash.update(rrsig.signer.to_digestable(origin))

        if rrsig.labels < len(rrname) - 1:
            suffix = rrname.split(rrsig.labels + 1)[1]
            rrname = dns.name.from_text('*', suffix)
        rrnamebuf = rrname.to_digestable(origin)
        rrfixed = struct.pack('!HHI', rdataset.rdtype, rdataset.rdclass,
                              rrsig.original_ttl)
        rrlist = sorted(rdataset)
        for rr in rrlist:
            hash.update(rrnamebuf)
            hash.update(rrfixed)
            rrdata = rr.to_digestable(origin)
            rrlen = struct.pack('!H', len(rrdata))
            hash.update(rrlen)
            hash.update(rrdata)

        try:
            if _is_rsa(rrsig.algorithm):
                verifier = pkcs1_15.new(pubkey)
                # will raise ValueError if verify fails:
                verifier.verify(hash, sig)
            elif _is_dsa(rrsig.algorithm):
                verifier = DSS.new(pubkey, 'fips-186-3')
                verifier.verify(hash, sig)
            elif _is_ecdsa(rrsig.algorithm):
                digest = hash.digest()
                if not pubkey.verify(digest, sig):
                    raise ValueError
            else:
                # Raise here for code clarity; this won't actually ever happen
                # since if the algorithm is really unknown we'd already have
                # raised an exception above
                raise ValidationFailure('unknown algorithm %u' % rrsig.algorithm)
            # If we got here, we successfully verified so we can return without error
            return
        except ValueError:
            # this happens on an individual validation failure
            continue
    # nothing verified -- raise failure:
    raise ValidationFailure('verify failure')


def _validate(rrset, rrsigset, keys, origin=None, now=None):
    """Validate an RRset.

    *rrset* is the RRset to validate.  It can be a ``dns.rrset.RRset`` or
    a ``(dns.name.Name, dns.rdataset.Rdataset)`` tuple.

    *rrsigset* is the signature RRset to be validated.  It can be a
    ``dns.rrset.RRset`` or a ``(dns.name.Name, dns.rdataset.Rdataset)`` tuple.

    *keys* is the key dictionary, used to find the DNSKEY associated with
    a given name.  The dictionary is keyed by a ``dns.name.Name``, and has
    ``dns.node.Node`` or ``dns.rdataset.Rdataset`` values.

    *origin* is a ``dns.name.Name``, the origin to use for relative names.

    *now* is an ``int``, the time to use when validating the signatures,
    in seconds since the UNIX epoch.  The default is the current time.
    """

    if isinstance(origin, string_types):
        origin = dns.name.from_text(origin, dns.name.root)

    if isinstance(rrset, tuple):
        rrname = rrset[0]
    else:
        rrname = rrset.name

    if isinstance(rrsigset, tuple):
        rrsigname = rrsigset[0]
        rrsigrdataset = rrsigset[1]
    else:
        rrsigname = rrsigset.name
        rrsigrdataset = rrsigset

    rrname = rrname.choose_relativity(origin)
    rrsigname = rrsigname.choose_relativity(origin)
    if rrname != rrsigname:
        raise ValidationFailure("owner names do not match")

    for rrsig in rrsigrdataset:
        try:
            _validate_rrsig(rrset, rrsig, keys, origin, now)
            return
        except ValidationFailure:
            pass
    raise ValidationFailure("no RRSIGs validated")


def _need_pycrypto(*args, **kwargs):
    raise NotImplementedError("DNSSEC validation requires pycryptodome/pycryptodomex")


try:
    try:
        # test we're using pycryptodome, not pycrypto (which misses SHA1 for example)
        from Crypto.Hash import MD5, SHA1, SHA256, SHA384, SHA512
        from Crypto.PublicKey import RSA as CryptoRSA, DSA as CryptoDSA
        from Crypto.Signature import pkcs1_15, DSS
        from Crypto.Util import number
    except ImportError:
        from Cryptodome.Hash import MD5, SHA1, SHA256, SHA384, SHA512
        from Cryptodome.PublicKey import RSA as CryptoRSA, DSA as CryptoDSA
        from Cryptodome.Signature import pkcs1_15, DSS
        from Cryptodome.Util import number
except ImportError:
    validate = _need_pycrypto
    validate_rrsig = _need_pycrypto
    _have_pycrypto = False
    _have_ecdsa = False
else:
    validate = _validate
    validate_rrsig = _validate_rrsig
    _have_pycrypto = True

    try:
        import ecdsa
        import ecdsa.ecdsa
        import ecdsa.ellipticcurve
        import ecdsa.keys
    except ImportError:
        _have_ecdsa = False
    else:
        _have_ecdsa = True

        class ECKeyWrapper(object):

            def __init__(self, key, key_len):
                self.key = key
                self.key_len = key_len

            def verify(self, digest, sig):
                diglong = number.bytes_to_long(digest)
                return self.key.pubkey.verifies(diglong, sig)



# -*- coding: utf-8 -*-

import json
import logging
from xml.etree.ElementTree import Element

from api.order import get_order_service
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)

from sysactions import (
    action,
    show_any_dialog,
    StopAction,
)

from actions.delivery.delivery_address_information import (
    insert_or_update_customer,
    insert_or_update_customer_with_confirmation,
)
from custom_sysactions import block_action_if_there_is_no_order
from order_api.dto import CustomProperty

logger = logging.getLogger("PosActions")


def handle_customer_data_on_total(
    pos_id,         # type: str
    order_picture,  # type: Element
):
    # type: (...) -> None

    if order_picture.attrib["saleTypeDescr"] != "DELIVERY":
        return

    from posactions import ask_customer_delivery_data
    if ask_customer_delivery_data != "onTotal":
        return

    phone_property = order_picture.find('.//OrderProperty[@key="CUSTOMER_PHONE"]')
    if phone_property is not None and phone_property.attrib["value"] != "":
        return

    customer_data = request_and_save_customer_data(pos_id=pos_id)
    if customer_data is None:
        raise StopAction()

    _add_customer_data_properties_on_order(pos_id, order_picture, customer_data)


@action
@block_action_if_there_is_no_order
def edit_order_delivery_address(
    pos_id,         # type: str
):
    # type: (...) -> None

    customer_data = request_and_save_customer_data(pos_id=pos_id)
    if customer_data is None:
        return

    order = get_order_service(pos_id=pos_id).get_order_picture()
    _add_customer_data_properties_on_order(
        pos_id=pos_id,
        order=order,
        customer_data=customer_data,
    )


@action
def save_delivery_address(
    pos_id,     # type: str
):
    # type: (...) -> None

    if get_order_service(pos_id=pos_id).has_current_order():
        edit_order_delivery_address(pos_id=pos_id)
        return

    selected_option, customer_json = full_request_customer_data(
        pos_id=pos_id,
        dialog_type="saveCustomerDeliveryAddress",
    )
    if selected_option is None:
        return

    insert_or_update_customer(
        customer_json=customer_json,
    )

    if selected_option == 2:
        from posactions import do_start_order
        do_start_order(
            pos_id=pos_id,
            sale_type="DELIVERY",
            customer_data=customer_json,
        )


def request_and_save_customer_data(
    pos_id,                           # type: str
):
    # type: (...) -> Optional[Dict[str, Any]]

    customer_json = request_customer_data(pos_id=pos_id)
    if not customer_json:
        return None

    insert_or_update_customer_with_confirmation(
        pos_id=pos_id,
        customer_json=customer_json,
    )

    return customer_json


def request_customer_data(
    pos_id,     # type: str
):
    # type: (...) -> Optional[Dict[str, str]]

    _, customer_json = full_request_customer_data(
        pos_id=pos_id,
        dialog_type="deliveryAddress",
    )

    return customer_json


def full_request_customer_data(
    pos_id,             # type: str
    dialog_type,        # type: str
):
    # type: (...) -> Tuple[Optional[int], Optional[Dict[str, str]]]

    customer = show_any_dialog(
        posid=pos_id,
        dlgtype=dialog_type,
        message="",
        title="",
        infotext="",
        buttons="",
        icon="",
        timeout=6000000,
        mask="",
        defvalue="",
        minvalue="",
        maxvalue="",
        contents="",
        asynch=False,
    )
    customer_json = json.loads(customer)

    if not customer_json:
        return None, None

    selected_option = customer_json["SELECTED_OPTION"]
    customer_json.pop('SELECTED_OPTION')
    return selected_option, customer_json


def _add_customer_data_properties_on_order(
    pos_id,             # type: str
    order,              # type: Element
    customer_data,      # type: Dict[str, Any]
):
    # type: (...) -> None

    properties_to_add = []
    properties_to_remove = []
    for key, value in customer_data.items():
        if value is None and order.find('.//OrderProperty[@key="{}"]'.format(key)) is not None:
            properties_to_remove.append(key)
        else:
            properties_to_add.append(CustomProperty(
                key=key,
                value=value,
            ))

    order_service = get_order_service(pos_id=pos_id)
    if properties_to_add:
        order_service.set_order_custom_properties(custom_properties=properties_to_add)

    if properties_to_remove:
        order_service.remove_order_custom_properties(keys=properties_to_remove)



import json
from typing import (
    Optional,
)

from xml.etree import cElementTree as eTree


def get_tender_detail(
    tender_line,  # type: eTree.Element
):
    # type: (...) -> Optional[str]

    tender_detail_json = tender_line.get("tenderDetail")
    if not tender_detail_json:
        return None

    try:
        tender_detail = json.loads(tender_detail_json)
    except TypeError:
        return None

    for value in tender_detail.keys():
        if tender_detail[value] is None:
            tender_detail[value] = ""
    return json.dumps(tender_detail)



from escposformatter import DefaultEscPosFormatter
from htmlformatter import HtmlFormatter
from report import (
    ReportParser,
    Report,
)
from sysactions import (
    generate_report,
    print_text,
    show_print_preview,
    show_messagebox,
)

from typing import (
    Any,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree

from enum import Enum

from utilfunctions import (
    get_device_service,
    get_sector_name,
)


class PrinterType(Enum):
    esc_pos = "ESC_POS"
    apos = "APOS"


def print_report(
    pos_id,               # type: Any[Union[str, int]]
    report_name,          # type: str
    report_args,          # type: list
    preview=True,         # type: Optional[bool]
    order=None,           # type: Optional[eTree]
    allow_printing=True,  # type: Optional[bool]
    allow_accents=False,  # type: Optional[bool]
    is_mobile=False,      # type: Optional[bool]
):
    # type: (...) -> Optional[bool]

    report_data = generate_report(
        report_name,
        pos_id,
        *report_args
    )
    report = ReportParser().parse(report=report_data)

    current_printer = _get_printer(
        pos_id=pos_id,
        order=order,
    )

    if report is not None:
        if preview:
            print_preview_report = HtmlFormatter().format(report)
            result = show_print_preview(
                posid=pos_id,
                text_data=print_preview_report.encode("utf-8"),
                title="$PRINT_PREVIEW",
                buttons="$PRINT|$CLOSE" if allow_printing else "$CLOSE",
            )
            if result in (1, None) or not allow_printing:
                return False

        print_data = _get_print_data(
            report=report,
            is_mobile=is_mobile,
            allow_accents=allow_accents,
        )

        print_text(
            posid=pos_id,
            model=None,
            text=print_data,
            force_printer=current_printer,
        )
    else:
        show_messagebox(
            posid=pos_id,
            title="$NO_REPORT_DATA",
            message="$ERROR",
        )


def _get_print_data(
    report,         # type: Report
    is_mobile,      # type: bool
    allow_accents,  # type: bool
):
    # type: (...) -> str

    if is_mobile:
        return DefaultEscPosFormatter().format_mobile(
            report=report,
        )

    return DefaultEscPosFormatter().format(
        report=report,
        allow_accents=allow_accents,
    )


def _get_printer(
    pos_id,         # type: Any[Union[str, int]]
    order=None,     # type: Optional[eTree]
):
    # type: (...) -> str

    sector_name = get_sector_name(order=order) if order else None
    current_printer = get_device_service(
        pos_id=pos_id,
        device_type="printer",
        sector=sector_name,
    )
    return current_printer



import hashlib
from typing import Optional

from ..totp import TOTP

STEAM_CHARS = "23456789BCDFGHJKMNPQRTVWXY"  # steam's custom alphabet
STEAM_DEFAULT_DIGITS = 5  # Steam TOTP code length


class Steam(TOTP):
    """
    Steam's custom TOTP. Subclass of `pyotp.totp.TOTP`.
    """

    def __init__(
        self,
        s: str,
        name: Optional[str] = None,
        issuer: Optional[str] = None,
        interval: int = 30,
        digits: int = 5
    ) -> None:
        """
        :param s: secret in base32 format
        :param interval: the time interval in seconds for OTP. This defaults to 30.
        :param name: account name
        :param issuer: issuer
        """
        self.interval = interval
        super().__init__(s=s, digits=10, digest=hashlib.sha1, name=name, issuer=issuer)

    def generate_otp(self, input: int) -> str:
        """
        :param input: the HMAC counter value to use as the OTP input.
            Usually either the counter, or the computed integer based on the Unix timestamp
        """
        str_code = super().generate_otp(input)
        int_code = int(str_code)

        steam_code = ""
        total_chars = len(STEAM_CHARS)

        for _ in range(STEAM_DEFAULT_DIGITS):
            pos = int_code % total_chars
            char = STEAM_CHARS[int(pos)]
            steam_code += char
            int_code //= total_chars

        return steam_code



# -*- coding: utf-8 -*-

from _OrderDataFormatter import OrderDataFormatter
from repository import OrderRetrieverRepository


class OrderRetrieverService(object):
    def __init__(self, order_retriever_repository, order_data_formatter):
        # type: (OrderRetrieverRepository, OrderDataFormatter) -> None
        self.order_retriever_repository = order_retriever_repository
        self.order_data_formatter = order_data_formatter

    def get_orders_data(self, data):
        # type: (str) -> unicode
        data_list = data.split("\0")
        initial_order_id = int(data_list[0])
        last_order_id = None
        if len(data_list) > 1:
            last_order_id = int(data_list[1])

        order_data_list = self.order_retriever_repository.get_order_data_list(initial_order_id, last_order_id)
        return self.order_data_formatter.format_order_data_list(order_data_list)

    def get_last_order_id(self):
        return str(self.order_retriever_repository.get_last_order_id())



from abc import (
    ABCMeta,
    abstractmethod,
)

from messagebus import (
    Message,
    MessageBus,
    Event,
    InternalErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor


class NoErrorProcessor(UuidMessageProcessor):

    __metaclass__ = ABCMeta

    @abstractmethod
    def parse_data(self, data, event_type):
        raise NotImplementedError()

    @abstractmethod
    def call_business(self, obj):
        raise NotImplementedError()

    @abstractmethod
    def format_response(self, message_bus, message, event, input_obj, result):
        raise NotImplementedError()

    def format_exception(
        self,           # type: NoErrorProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: str
        exception,      # type: Exception
    ):
        # type: (...) -> None

        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(exception),
        )

    def format_parse_exception(
        self,           # type: NoErrorProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: bytes
        exception,      # type: Exception
    ):
        # type: (...) -> None

        self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception,
        )



# util/_concurrency_py3k.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: allow-untyped-defs, allow-untyped-calls

from __future__ import annotations

import asyncio
from contextvars import Context
import sys
import typing
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Coroutine
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

from .langhelpers import memoized_property
from .. import exc
from ..util.typing import Protocol
from ..util.typing import TypeGuard

_T = TypeVar("_T")

if typing.TYPE_CHECKING:

    class greenlet(Protocol):
        dead: bool
        gr_context: Optional[Context]

        def __init__(self, fn: Callable[..., Any], driver: greenlet):
            ...

        def throw(self, *arg: Any) -> Any:
            return None

        def switch(self, value: Any) -> Any:
            return None

    def getcurrent() -> greenlet:
        ...

else:
    from greenlet import getcurrent
    from greenlet import greenlet


# If greenlet.gr_context is present in current version of greenlet,
# it will be set with the current context on creation.
# Refs: https://github.com/python-greenlet/greenlet/pull/198
_has_gr_context = hasattr(getcurrent(), "gr_context")


def is_exit_exception(e: BaseException) -> bool:
    # note asyncio.CancelledError is already BaseException
    # so was an exit exception in any case
    return not isinstance(e, Exception) or isinstance(
        e, (asyncio.TimeoutError, asyncio.CancelledError)
    )


# implementation based on snaury gist at
# https://gist.github.com/snaury/202bf4f22c41ca34e56297bae5f33fef
# Issue for context: https://github.com/python-greenlet/greenlet/issues/173


class _AsyncIoGreenlet(greenlet):
    dead: bool

    def __init__(self, fn: Callable[..., Any], driver: greenlet):
        greenlet.__init__(self, fn, driver)
        self.driver = driver
        if _has_gr_context:
            self.gr_context = driver.gr_context


_T_co = TypeVar("_T_co", covariant=True)

if TYPE_CHECKING:

    def iscoroutine(
        awaitable: Awaitable[_T_co],
    ) -> TypeGuard[Coroutine[Any, Any, _T_co]]:
        ...

else:
    iscoroutine = asyncio.iscoroutine


def _safe_cancel_awaitable(awaitable: Awaitable[Any]) -> None:
    # https://docs.python.org/3/reference/datamodel.html#coroutine.close

    if iscoroutine(awaitable):
        awaitable.close()


def await_only(awaitable: Awaitable[_T]) -> _T:
    """Awaits an async function in a sync method.

    The sync method must be inside a :func:`greenlet_spawn` context.
    :func:`await_only` calls cannot be nested.

    :param awaitable: The coroutine to call.

    """
    # this is called in the context greenlet while running fn
    current = getcurrent()
    if not isinstance(current, _AsyncIoGreenlet):
        _safe_cancel_awaitable(awaitable)

        raise exc.MissingGreenlet(
            "greenlet_spawn has not been called; can't call await_only() "
            "here. Was IO attempted in an unexpected place?"
        )

    # returns the control to the driver greenlet passing it
    # a coroutine to run. Once the awaitable is done, the driver greenlet
    # switches back to this greenlet with the result of awaitable that is
    # then returned to the caller (or raised as error)
    return current.driver.switch(awaitable)  # type: ignore[no-any-return]


def await_fallback(awaitable: Awaitable[_T]) -> _T:
    """Awaits an async function in a sync method.

    The sync method must be inside a :func:`greenlet_spawn` context.
    :func:`await_fallback` calls cannot be nested.

    :param awaitable: The coroutine to call.

    """

    # this is called in the context greenlet while running fn
    current = getcurrent()
    if not isinstance(current, _AsyncIoGreenlet):
        loop = get_event_loop()
        if loop.is_running():
            _safe_cancel_awaitable(awaitable)

            raise exc.MissingGreenlet(
                "greenlet_spawn has not been called and asyncio event "
                "loop is already running; can't call await_fallback() here. "
                "Was IO attempted in an unexpected place?"
            )
        return loop.run_until_complete(awaitable)

    return current.driver.switch(awaitable)  # type: ignore[no-any-return]


async def greenlet_spawn(
    fn: Callable[..., _T],
    *args: Any,
    _require_await: bool = False,
    **kwargs: Any,
) -> _T:
    """Runs a sync function ``fn`` in a new greenlet.

    The sync function can then use :func:`await_only` to wait for async
    functions.

    :param fn: The sync callable to call.
    :param \\*args: Positional arguments to pass to the ``fn`` callable.
    :param \\*\\*kwargs: Keyword arguments to pass to the ``fn`` callable.
    """

    result: Any
    context = _AsyncIoGreenlet(fn, getcurrent())
    # runs the function synchronously in gl greenlet. If the execution
    # is interrupted by await_only, context is not dead and result is a
    # coroutine to wait. If the context is dead the function has
    # returned, and its result can be returned.
    switch_occurred = False
    try:
        result = context.switch(*args, **kwargs)
        while not context.dead:
            switch_occurred = True
            try:
                # wait for a coroutine from await_only and then return its
                # result back to it.
                value = await result
            except BaseException:
                # this allows an exception to be raised within
                # the moderated greenlet so that it can continue
                # its expected flow.
                result = context.throw(*sys.exc_info())
            else:
                result = context.switch(value)
    finally:
        # clean up to avoid cycle resolution by gc
        del context.driver
    if _require_await and not switch_occurred:
        raise exc.AwaitRequired(
            "The current operation required an async execution but none was "
            "detected. This will usually happen when using a non compatible "
            "DBAPI driver. Please ensure that an async DBAPI is used."
        )
    return result  # type: ignore[no-any-return]


class AsyncAdaptedLock:
    @memoized_property
    def mutex(self) -> asyncio.Lock:
        # there should not be a race here for coroutines creating the
        # new lock as we are not using await, so therefore no concurrency
        return asyncio.Lock()

    def __enter__(self) -> bool:
        # await is used to acquire the lock only after the first calling
        # coroutine has created the mutex.
        return await_fallback(self.mutex.acquire())

    def __exit__(self, *arg: Any, **kw: Any) -> None:
        self.mutex.release()


def _util_async_run_coroutine_function(
    fn: Callable[..., Coroutine[Any, Any, Any]], *args: Any, **kwargs: Any
) -> Any:
    """for test suite/ util only"""

    loop = get_event_loop()
    if loop.is_running():
        raise Exception(
            "for async run coroutine we expect that no greenlet or event "
            "loop is running when we start out"
        )
    return loop.run_until_complete(fn(*args, **kwargs))


def _util_async_run(
    fn: Callable[..., Coroutine[Any, Any, Any]], *args: Any, **kwargs: Any
) -> Any:
    """for test suite/ util only"""

    loop = get_event_loop()
    if not loop.is_running():
        return loop.run_until_complete(greenlet_spawn(fn, *args, **kwargs))
    else:
        # allow for a wrapped test function to call another
        assert isinstance(getcurrent(), _AsyncIoGreenlet)
        return fn(*args, **kwargs)


def get_event_loop() -> asyncio.AbstractEventLoop:
    """vendor asyncio.get_event_loop() for python 3.7 and above.

    Python 3.10 deprecates get_event_loop() as a standalone.

    """
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        # avoid "During handling of the above exception, another exception..."
        pass
    return asyncio.get_event_loop_policy().get_event_loop()



from pos_model import Order
from typing import (
    Dict,
    Any,
)

from _ContextKeys import ContextKeys
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from common import InfAdicBuilderCommon
from ibpttaxprocessor import IbptTaxProcessor
from models import NfeTender
from services import TPagService


class InfAdicBuilder(NfceXmlPartBuilder):

    def __init__(
        self,                       # type: str
        customer_text_field,        # type: str
        tax_builder,                # type: IbptTaxProcessor
        coupon_fiscal_parameters,   # type: Dict[str, bool]
        tenders_type,               # type: Dict[str, Any]
    ):
        # type: (...) -> None

        self.customer_text_field = customer_text_field
        self.tax_builder = tax_builder
        self.coupon_fiscal_parameters = coupon_fiscal_parameters
        self.tenders_type = tenders_type

    def build_xml(
        self,       # type: InfAdicBuilder
        order,      # type: Order
        context,    # type: ContextKeys
    ):
        # type: (...) -> unicode

        inf_adic_builder_common = InfAdicBuilderCommon(
            order=order,
            customer_text_field=self.customer_text_field,
            tax_builder=self.tax_builder,
            coupon_fiscal_parameters=self.coupon_fiscal_parameters,
        )

        return u"<infAdic><infCpl>{0}</infCpl>{1}</infAdic>".format(
            inf_adic_builder_common.get_inf_cpl(),
            self._get_tenders_additional_info(order=order),
        )

    def _get_tenders_additional_info(
        self,     # type: InfAdicBuilder
        order,    # type: Order
    ):
        # type: (...) -> unicode

        xml = u""
        is_static_pix_already_added = False
        t_pag_service = TPagService(tenders_type=self.tenders_type)
        for tender in order.tenders:
            t_pag = t_pag_service.get_t_pag(tender=tender)
            if t_pag == NfeTender.pix_estatico and not is_static_pix_already_added:
                xml += u"""<obsCont xCampo="tPix"><xTexto>1</xTexto></obsCont>"""
                is_static_pix_already_added = True
            elif t_pag == NfeTender.pagamento_instantaneo_pix:
                xml += u"""<obsCont xCampo="tPix"><xTexto>0</xTexto></obsCont>"""
                xml += u"""<obsCont xCampo="idPIX"><xTexto>{}</xTexto></obsCont>""".format(
                    tender.nsu_host or tender.auth_code
                )

        return xml



from messagebus import TokenCreator, TokenPriority

POS_CONTROL_GROUP = "5"

TK_POS_GET_MODEL = TokenCreator.create_token(
    priority=TokenPriority.high,
    message_group=POS_CONTROL_GROUP,
    number="1"
)

TK_POS_GETCONFIG = TokenCreator.create_token(
    priority=TokenPriority.high,
    message_group=POS_CONTROL_GROUP,
    number="16",
)



class ProductBarcodeNotFound(Exception):

    def __init__(
        self,       # type: ProductBarcodeNotFound
        part_code,  # type: str
    ):
        # type: (...) -> None
        super(ProductBarcodeNotFound, self).__init__(part_code)

        self.part_code = part_code



import json
from typing import Dict

from nreports.interactor.build_report import BuildReportInteractor
from nreports.interactor.build_report.dto import BuildReportRequest
from nreports_api.shared.report_dto import ReportDto
from ._BaseNReportProcessor import BaseNReportProcessor


class BuildReportProcessor(BaseNReportProcessor):

    def __init__(
        self,
        interactor: BuildReportInteractor,
    ) -> None:
        super().__init__()

        self.interactor = interactor

    def parse_data(
        self,
        data: bytes,
    ) -> BuildReportRequest:

        try:
            data_dict = json.loads(s=data.decode())
        except Exception:
            raise self.InvalidJson()

        report_name = self._parse_report_name(data_dict=data_dict)

        return BuildReportRequest(
            pos_id=data_dict["posId"],
            report_name=report_name,
            data=data_dict,
        )

    def call_business(
        self,
        request: BuildReportRequest,
    ) -> ReportDto:

        return self.interactor.execute(request=request)

    def _parse_report_name(
        self,
        data_dict: Dict,
    ) -> str:

        report_name = None
        try:
            report_name = str(data_dict["reportName"].strip())
        finally:
            if not report_name:
                raise self.MissingReportName()

        return report_name



import json
import logging

from application.interactor.get_loyalty_customer import GetLoyaltyCustomerInfoInteractor
from application.interactor.get_loyalty_customer.dto import GetLoyaltyCustomerInfoRequest
from application.model import AsyncUuidMessageProcessor
from application.model.customexception import NotFoundError
from messagebus import (
    DataType,
    Message,
    AckMessage,
    MessageBus,
    Event,
    InternalErrorMessage,
)
from .tokens import ResponseToken

logger = logging.getLogger(__name__)


class GetLoyaltyCustomerInfoProcessor(AsyncUuidMessageProcessor):
    def __init__(
        self,        # type: GetLoyaltyCustomerInfoProcessor
        interactor,  # type: GetLoyaltyCustomerInfoInteractor
    ):
        # type: (...) -> None

        super(GetLoyaltyCustomerInfoProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,        # type: GetLoyaltyCustomerInfoProcessor
        data,        # type: bytes
        event_type,  # type: str
    ):
        # type: (...) -> GetLoyaltyCustomerInfoRequest

        request = json.loads(data.decode("utf-8"))

        return GetLoyaltyCustomerInfoRequest(
            pos_id=int(request["posId"]),
            loyalty_id=request["loyaltyId"],
            order_id=request["orderId"] or None,
            table_id=request["tableId"] or None,
            seat=request["seat"],
        )

    def call_business(
        self,       # type: GetLoyaltyCustomerInfoProcessor
        request,    # type: GetLoyaltyCustomerInfoRequest
    ):
        # type: (...) -> None

        return self.interactor.get_loyalty_customer_info(request=request)

    def format_response(
        self,         # type: GetLoyaltyCustomerInfoProcessor
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: Event
        request,      # type: str
        response,     # type: str
    ):
        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(
                data_type=DataType.string,
                data=response
            )
        )

    def format_exception(
        self,
        message_bus,  # type: MessageBus
        message,      # type: Message
        event,        # type: Event
        input_obj,    # type: str
        exception,    # type: Exception
    ):
        if isinstance(exception, NotFoundError):
            message_bus.reply_message(
                message=message,
                reply_message=Message(token=ResponseToken.TK_LOYALTY_ID_NOT_FOUND.value)
            )
        else:
            message_bus.reply_message(
                message=message,
                reply_message=InternalErrorMessage(exception)
            )

    def format_parse_exception(
        self,
        message_bus,   # type: MessageBus
        message,       # type: Message
        event,         # type: Event
        data,          # type: bytes
        exception,     # type: Exception
    ):
        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(exception)
        )



from dataclasses import (
    dataclass,
    field,
)
from typing import (
    List,
    Optional,
)

from boss.model import User


@dataclass
class Configs:

    @dataclass
    class ApiRetryPolicy:
        count: int = 1
        sleep: int = 1

    @dataclass
    class PumpCredentials:
        base_url: str = ""
        api_key: str = ""

    @dataclass
    class Endpoints:
        authenticate: str = ""

    otp_interval: int = 600
    database_path: str = ""
    required_services: str = ""
    api_retry_policy: Optional[ApiRetryPolicy] = None
    endpoints: Optional[Endpoints] = None
    pump_credentials: Optional[PumpCredentials] = None
    store_id: Optional[int] = None
    default_users: List[User] = field(default_factory=list)
    component_name: str = "boss"
    service_name: str = "UserControl"



# orm/dynamic.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php


"""Dynamic collection API.

Dynamic collections act like Query() objects for read operations and support
basic add/delete mutation.

.. legacy:: the "dynamic" loader is a legacy feature, superseded by the
 "write_only" loader.


"""

from __future__ import annotations

from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from . import attributes
from . import exc as orm_exc
from . import relationships
from . import util as orm_util
from .base import PassiveFlag
from .query import Query
from .session import object_session
from .writeonly import AbstractCollectionWriter
from .writeonly import WriteOnlyAttributeImpl
from .writeonly import WriteOnlyHistory
from .writeonly import WriteOnlyLoader
from .. import util
from ..engine import result


if TYPE_CHECKING:
    from . import QueryableAttribute
    from .mapper import Mapper
    from .relationships import _RelationshipOrderByArg
    from .session import Session
    from .state import InstanceState
    from .util import AliasedClass
    from ..event import _Dispatch
    from ..sql.elements import ColumnElement

_T = TypeVar("_T", bound=Any)


class DynamicCollectionHistory(WriteOnlyHistory[_T]):
    def __init__(
        self,
        attr: DynamicAttributeImpl,
        state: InstanceState[_T],
        passive: PassiveFlag,
        apply_to: Optional[DynamicCollectionHistory[_T]] = None,
    ) -> None:
        if apply_to:
            coll = AppenderQuery(attr, state).autoflush(False)
            self.unchanged_items = util.OrderedIdentitySet(coll)
            self.added_items = apply_to.added_items
            self.deleted_items = apply_to.deleted_items
            self._reconcile_collection = True
        else:
            self.deleted_items = util.OrderedIdentitySet()
            self.added_items = util.OrderedIdentitySet()
            self.unchanged_items = util.OrderedIdentitySet()
            self._reconcile_collection = False


class DynamicAttributeImpl(WriteOnlyAttributeImpl):
    _supports_dynamic_iteration = True
    collection_history_cls = DynamicCollectionHistory[Any]
    query_class: Type[AppenderMixin[Any]]  # type: ignore[assignment]

    def __init__(
        self,
        class_: Union[Type[Any], AliasedClass[Any]],
        key: str,
        dispatch: _Dispatch[QueryableAttribute[Any]],
        target_mapper: Mapper[_T],
        order_by: _RelationshipOrderByArg,
        query_class: Optional[Type[AppenderMixin[_T]]] = None,
        **kw: Any,
    ) -> None:
        attributes.AttributeImpl.__init__(
            self, class_, key, None, dispatch, **kw
        )
        self.target_mapper = target_mapper
        if order_by:
            self.order_by = tuple(order_by)
        if not query_class:
            self.query_class = AppenderQuery
        elif AppenderMixin in query_class.mro():
            self.query_class = query_class
        else:
            self.query_class = mixin_user_query(query_class)


@relationships.RelationshipProperty.strategy_for(lazy="dynamic")
class DynaLoader(WriteOnlyLoader):
    impl_class = DynamicAttributeImpl


class AppenderMixin(AbstractCollectionWriter[_T]):
    """A mixin that expects to be mixing in a Query class with
    AbstractAppender.


    """

    query_class: Optional[Type[Query[_T]]] = None
    _order_by_clauses: Tuple[ColumnElement[Any], ...]

    def __init__(
        self, attr: DynamicAttributeImpl, state: InstanceState[_T]
    ) -> None:
        Query.__init__(
            self,  # type: ignore[arg-type]
            attr.target_mapper,
            None,
        )
        super().__init__(attr, state)

    @property
    def session(self) -> Optional[Session]:
        sess = object_session(self.instance)
        if sess is not None and sess.autoflush and self.instance in sess:
            sess.flush()
        if not orm_util.has_identity(self.instance):
            return None
        else:
            return sess

    @session.setter
    def session(self, session: Session) -> None:
        self.sess = session

    def _iter(self) -> Union[result.ScalarResult[_T], result.Result[_T]]:
        sess = self.session
        if sess is None:
            state = attributes.instance_state(self.instance)
            if state.detached:
                util.warn(
                    "Instance %s is detached, dynamic relationship cannot "
                    "return a correct result.   This warning will become "
                    "a DetachedInstanceError in a future release."
                    % (orm_util.state_str(state))
                )

            return result.IteratorResult(
                result.SimpleResultMetaData([self.attr.class_.__name__]),
                self.attr._get_collection_history(  # type: ignore[arg-type]
                    attributes.instance_state(self.instance),
                    PassiveFlag.PASSIVE_NO_INITIALIZE,
                ).added_items,
                _source_supports_scalars=True,
            ).scalars()
        else:
            return self._generate(sess)._iter()

    if TYPE_CHECKING:

        def __iter__(self) -> Iterator[_T]:
            ...

    def __getitem__(self, index: Any) -> Union[_T, List[_T]]:
        sess = self.session
        if sess is None:
            return self.attr._get_collection_history(
                attributes.instance_state(self.instance),
                PassiveFlag.PASSIVE_NO_INITIALIZE,
            ).indexed(index)
        else:
            return self._generate(sess).__getitem__(index)  # type: ignore[no-any-return] # noqa: E501

    def count(self) -> int:
        sess = self.session
        if sess is None:
            return len(
                self.attr._get_collection_history(
                    attributes.instance_state(self.instance),
                    PassiveFlag.PASSIVE_NO_INITIALIZE,
                ).added_items
            )
        else:
            return self._generate(sess).count()

    def _generate(
        self,
        sess: Optional[Session] = None,
    ) -> Query[_T]:
        # note we're returning an entirely new Query class instance
        # here without any assignment capabilities; the class of this
        # query is determined by the session.
        instance = self.instance
        if sess is None:
            sess = object_session(instance)
            if sess is None:
                raise orm_exc.DetachedInstanceError(
                    "Parent instance %s is not bound to a Session, and no "
                    "contextual session is established; lazy load operation "
                    "of attribute '%s' cannot proceed"
                    % (orm_util.instance_str(instance), self.attr.key)
                )

        if self.query_class:
            query = self.query_class(self.attr.target_mapper, session=sess)
        else:
            query = sess.query(self.attr.target_mapper)

        query._where_criteria = self._where_criteria
        query._from_obj = self._from_obj
        query._order_by_clauses = self._order_by_clauses

        return query

    def add_all(self, iterator: Iterable[_T]) -> None:
        """Add an iterable of items to this :class:`_orm.AppenderQuery`.

        The given items will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        This method is provided to assist in delivering forwards-compatibility
        with the :class:`_orm.WriteOnlyCollection` collection class.

        .. versionadded:: 2.0

        """
        self._add_all_impl(iterator)

    def add(self, item: _T) -> None:
        """Add an item to this :class:`_orm.AppenderQuery`.

        The given item will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        This method is provided to assist in delivering forwards-compatibility
        with the :class:`_orm.WriteOnlyCollection` collection class.

        .. versionadded:: 2.0

        """
        self._add_all_impl([item])

    def extend(self, iterator: Iterable[_T]) -> None:
        """Add an iterable of items to this :class:`_orm.AppenderQuery`.

        The given items will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        """
        self._add_all_impl(iterator)

    def append(self, item: _T) -> None:
        """Append an item to this :class:`_orm.AppenderQuery`.

        The given item will be persisted to the database in terms of
        the parent instance's collection on the next flush.

        """
        self._add_all_impl([item])

    def remove(self, item: _T) -> None:
        """Remove an item from this :class:`_orm.AppenderQuery`.

        The given item will be removed from the parent instance's collection on
        the next flush.

        """
        self._remove_impl(item)


class AppenderQuery(AppenderMixin[_T], Query[_T]):  # type: ignore[misc]
    """A dynamic query that supports basic collection storage operations.

    Methods on :class:`.AppenderQuery` include all methods of
    :class:`_orm.Query`, plus additional methods used for collection
    persistence.


    """


def mixin_user_query(cls: Any) -> type[AppenderMixin[Any]]:
    """Return a new class with AppenderQuery functionality layered over."""
    name = "Appender" + cls.__name__
    return type(name, (AppenderMixin, cls), {"query_class": cls})



import iso8601
from _StatisticBox import StatisticBox
from production.model import ProductionOrder
from typing import List, Any


class MeanTimeStatisticBox(StatisticBox):
    def __init__(self, name, sons, publish_scheduler, logger, get_config):
        super(MeanTimeStatisticBox, self).__init__(
            name,
            sons,
            publish_scheduler,
            logger,
            get_config
        )

        self.statistics = {}

    def get_order_value(self, order):
        # type: (ProductionOrder) -> Any

        start_time = iso8601.parse_date(order.tags[self.start_tag])
        end_time = iso8601.parse_date(order.tags[self.end_tag])
        return (end_time - start_time).total_seconds()

    def update_statistics(self, values):
        # type: (List) -> float

        return sum(values) / (len(values) or 1)



# testing/util.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors


from __future__ import annotations

from collections import deque
import decimal
import gc
from itertools import chain
import random
import sys
from sys import getsizeof
import types

from . import config
from . import mock
from .. import inspect
from ..engine import Connection
from ..schema import Column
from ..schema import DropConstraint
from ..schema import DropTable
from ..schema import ForeignKeyConstraint
from ..schema import MetaData
from ..schema import Table
from ..sql import schema
from ..sql.sqltypes import Integer
from ..util import decorator
from ..util import defaultdict
from ..util import has_refcount_gc
from ..util import inspect_getfullargspec


if not has_refcount_gc:

    def non_refcount_gc_collect(*args):
        gc.collect()
        gc.collect()

    gc_collect = lazy_gc = non_refcount_gc_collect
else:
    # assume CPython - straight gc.collect, lazy_gc() is a pass
    gc_collect = gc.collect

    def lazy_gc():
        pass


def picklers():
    picklers = set()
    import pickle

    picklers.add(pickle)

    # yes, this thing needs this much testing
    for pickle_ in picklers:
        for protocol in range(-2, pickle.HIGHEST_PROTOCOL + 1):
            yield pickle_.loads, lambda d: pickle_.dumps(d, protocol)


def random_choices(population, k=1):
    return random.choices(population, k=k)


def round_decimal(value, prec):
    if isinstance(value, float):
        return round(value, prec)

    # can also use shift() here but that is 2.6 only
    return (value * decimal.Decimal("1" + "0" * prec)).to_integral(
        decimal.ROUND_FLOOR
    ) / pow(10, prec)


class RandomSet(set):
    def __iter__(self):
        l = list(set.__iter__(self))
        random.shuffle(l)
        return iter(l)

    def pop(self):
        index = random.randint(0, len(self) - 1)
        item = list(set.__iter__(self))[index]
        self.remove(item)
        return item

    def union(self, other):
        return RandomSet(set.union(self, other))

    def difference(self, other):
        return RandomSet(set.difference(self, other))

    def intersection(self, other):
        return RandomSet(set.intersection(self, other))

    def copy(self):
        return RandomSet(self)


def conforms_partial_ordering(tuples, sorted_elements):
    """True if the given sorting conforms to the given partial ordering."""

    deps = defaultdict(set)
    for parent, child in tuples:
        deps[parent].add(child)
    for i, node in enumerate(sorted_elements):
        for n in sorted_elements[i:]:
            if node in deps[n]:
                return False
    else:
        return True


def all_partial_orderings(tuples, elements):
    edges = defaultdict(set)
    for parent, child in tuples:
        edges[child].add(parent)

    def _all_orderings(elements):
        if len(elements) == 1:
            yield list(elements)
        else:
            for elem in elements:
                subset = set(elements).difference([elem])
                if not subset.intersection(edges[elem]):
                    for sub_ordering in _all_orderings(subset):
                        yield [elem] + sub_ordering

    return iter(_all_orderings(elements))


def function_named(fn, name):
    """Return a function with a given __name__.

    Will assign to __name__ and return the original function if possible on
    the Python implementation, otherwise a new function will be constructed.

    This function should be phased out as much as possible
    in favor of @decorator.   Tests that "generate" many named tests
    should be modernized.

    """
    try:
        fn.__name__ = name
    except TypeError:
        fn = types.FunctionType(
            fn.__code__, fn.__globals__, name, fn.__defaults__, fn.__closure__
        )
    return fn


def run_as_contextmanager(ctx, fn, *arg, **kw):
    """Run the given function under the given contextmanager,
    simulating the behavior of 'with' to support older
    Python versions.

    This is not necessary anymore as we have placed 2.6
    as minimum Python version, however some tests are still using
    this structure.

    """

    obj = ctx.__enter__()
    try:
        result = fn(obj, *arg, **kw)
        ctx.__exit__(None, None, None)
        return result
    except:
        exc_info = sys.exc_info()
        raise_ = ctx.__exit__(*exc_info)
        if not raise_:
            raise
        else:
            return raise_


def rowset(results):
    """Converts the results of sql execution into a plain set of column tuples.

    Useful for asserting the results of an unordered query.
    """

    return {tuple(row) for row in results}


def fail(msg):
    assert False, msg


@decorator
def provide_metadata(fn, *args, **kw):
    """Provide bound MetaData for a single test, dropping afterwards.

    Legacy; use the "metadata" pytest fixture.

    """

    from . import fixtures

    metadata = schema.MetaData()
    self = args[0]
    prev_meta = getattr(self, "metadata", None)
    self.metadata = metadata
    try:
        return fn(*args, **kw)
    finally:
        # close out some things that get in the way of dropping tables.
        # when using the "metadata" fixture, there is a set ordering
        # of things that makes sure things are cleaned up in order, however
        # the simple "decorator" nature of this legacy function means
        # we have to hardcode some of that cleanup ahead of time.

        # close ORM sessions
        fixtures.close_all_sessions()

        # integrate with the "connection" fixture as there are many
        # tests where it is used along with provide_metadata
        cfc = fixtures.base._connection_fixture_connection
        if cfc:
            # TODO: this warning can be used to find all the places
            # this is used with connection fixture
            # warn("mixing legacy provide metadata with connection fixture")
            drop_all_tables_from_metadata(metadata, cfc)
            # as the provide_metadata fixture is often used with "testing.db",
            # when we do the drop we have to commit the transaction so that
            # the DB is actually updated as the CREATE would have been
            # committed
            cfc.get_transaction().commit()
        else:
            drop_all_tables_from_metadata(metadata, config.db)
        self.metadata = prev_meta


def flag_combinations(*combinations):
    """A facade around @testing.combinations() oriented towards boolean
    keyword-based arguments.

    Basically generates a nice looking identifier based on the keywords
    and also sets up the argument names.

    E.g.::

        @testing.flag_combinations(
            dict(lazy=False, passive=False),
            dict(lazy=True, passive=False),
            dict(lazy=False, passive=True),
            dict(lazy=False, passive=True, raiseload=True),
        )


    would result in::

        @testing.combinations(
            ('', False, False, False),
            ('lazy', True, False, False),
            ('lazy_passive', True, True, False),
            ('lazy_passive', True, True, True),
            id_='iaaa',
            argnames='lazy,passive,raiseload'
        )

    """

    keys = set()

    for d in combinations:
        keys.update(d)

    keys = sorted(keys)

    return config.combinations(
        *[
            ("_".join(k for k in keys if d.get(k, False)),)
            + tuple(d.get(k, False) for k in keys)
            for d in combinations
        ],
        id_="i" + ("a" * len(keys)),
        argnames=",".join(keys),
    )


def lambda_combinations(lambda_arg_sets, **kw):
    args = inspect_getfullargspec(lambda_arg_sets)

    arg_sets = lambda_arg_sets(*[mock.Mock() for arg in args[0]])

    def create_fixture(pos):
        def fixture(**kw):
            return lambda_arg_sets(**kw)[pos]

        fixture.__name__ = "fixture_%3.3d" % pos
        return fixture

    return config.combinations(
        *[(create_fixture(i),) for i in range(len(arg_sets))], **kw
    )


def resolve_lambda(__fn, **kw):
    """Given a no-arg lambda and a namespace, return a new lambda that
    has all the values filled in.

    This is used so that we can have module-level fixtures that
    refer to instance-level variables using lambdas.

    """

    pos_args = inspect_getfullargspec(__fn)[0]
    pass_pos_args = {arg: kw.pop(arg) for arg in pos_args}
    glb = dict(__fn.__globals__)
    glb.update(kw)
    new_fn = types.FunctionType(__fn.__code__, glb)
    return new_fn(**pass_pos_args)


def metadata_fixture(ddl="function"):
    """Provide MetaData for a pytest fixture."""

    def decorate(fn):
        def run_ddl(self):
            metadata = self.metadata = schema.MetaData()
            try:
                result = fn(self, metadata)
                metadata.create_all(config.db)
                # TODO:
                # somehow get a per-function dml erase fixture here
                yield result
            finally:
                metadata.drop_all(config.db)

        return config.fixture(scope=ddl)(run_ddl)

    return decorate


def force_drop_names(*names):
    """Force the given table names to be dropped after test complete,
    isolating for foreign key cycles

    """

    @decorator
    def go(fn, *args, **kw):
        try:
            return fn(*args, **kw)
        finally:
            drop_all_tables(config.db, inspect(config.db), include_names=names)

    return go


class adict(dict):
    """Dict keys available as attributes.  Shadows."""

    def __getattribute__(self, key):
        try:
            return self[key]
        except KeyError:
            return dict.__getattribute__(self, key)

    def __call__(self, *keys):
        return tuple([self[key] for key in keys])

    get_all = __call__


def drop_all_tables_from_metadata(metadata, engine_or_connection):
    from . import engines

    def go(connection):
        engines.testing_reaper.prepare_for_drop_tables(connection)

        if not connection.dialect.supports_alter:
            from . import assertions

            with assertions.expect_warnings(
                "Can't sort tables", assert_=False
            ):
                metadata.drop_all(connection)
        else:
            metadata.drop_all(connection)

    if not isinstance(engine_or_connection, Connection):
        with engine_or_connection.begin() as connection:
            go(connection)
    else:
        go(engine_or_connection)


def drop_all_tables(
    engine,
    inspector,
    schema=None,
    consider_schemas=(None,),
    include_names=None,
):
    if include_names is not None:
        include_names = set(include_names)

    if schema is not None:
        assert consider_schemas == (
            None,
        ), "consider_schemas and schema are mutually exclusive"
        consider_schemas = (schema,)

    with engine.begin() as conn:
        for table_key, fkcs in reversed(
            inspector.sort_tables_on_foreign_key_dependency(
                consider_schemas=consider_schemas
            )
        ):
            if table_key:
                if (
                    include_names is not None
                    and table_key[1] not in include_names
                ):
                    continue
                conn.execute(
                    DropTable(
                        Table(table_key[1], MetaData(), schema=table_key[0])
                    )
                )
            elif fkcs:
                if not engine.dialect.supports_alter:
                    continue
                for t_key, fkc in fkcs:
                    if (
                        include_names is not None
                        and t_key[1] not in include_names
                    ):
                        continue
                    tb = Table(
                        t_key[1],
                        MetaData(),
                        Column("x", Integer),
                        Column("y", Integer),
                        schema=t_key[0],
                    )
                    conn.execute(
                        DropConstraint(
                            ForeignKeyConstraint([tb.c.x], [tb.c.y], name=fkc)
                        )
                    )


def teardown_events(event_cls):
    @decorator
    def decorate(fn, *arg, **kw):
        try:
            return fn(*arg, **kw)
        finally:
            event_cls._clear()

    return decorate


def total_size(o):
    """Returns the approximate memory footprint an object and all of its
    contents.

    source: https://code.activestate.com/recipes/577504/


    """

    def dict_handler(d):
        return chain.from_iterable(d.items())

    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    seen = set()  # track which object id's have already been seen
    default_size = getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


def count_cache_key_tuples(tup):
    """given a cache key tuple, counts how many instances of actual
    tuples are found.

    used to alert large jumps in cache key complexity.

    """
    stack = [tup]

    sentinel = object()
    num_elements = 0

    while stack:
        elem = stack.pop(0)
        if elem is sentinel:
            num_elements += 1
        elif isinstance(elem, tuple):
            if elem:
                stack = list(elem) + [sentinel] + stack
    return num_elements



class User(object):
    def __init__(
        self,       # type: User
        id,         # type: int
        name,       # type: str
        level,      # typer: int
    ):
        self.id = id
        self.name = name
        self.level = level



from threading import (
    Condition,
)
from typing import (
    Generic,
    Callable,
    Sequence,
    Optional,
    TypeVar,
    List,
)


T = TypeVar("T")


class InFlightSyncedList(Generic[T]):
    def __init__(self, wait_condition: Condition):
        self.values: List[T] = []
        self.in_flight = {}

        self.stopped = False
        self.wait_condition = wait_condition

    def append(self, value: T):
        with self.wait_condition:
            self.values.append(value)
            self.wait_condition.notify_all()

    def extend(self, values: Sequence[T]):
        with self.wait_condition:
            for value in values:
                self.values.append(value)

    def retrieve(self, func: Callable[[T], bool]) -> Sequence[T]:
        with self.wait_condition:
            all_values = tuple(filter(func, self.values))
            ret = []
            for value in all_values:
                if value not in self.in_flight:
                    self.in_flight[value] = True
                    ret.append(value)

            return ret

    def first(self, func: Callable[[T], bool]) -> Optional[T]:
        with self.wait_condition:
            for value in self.values:
                if value not in self.in_flight and func(value):
                    self.in_flight[value] = True
                    return value

            return None

    def get_not_in_flight(self) -> Sequence[T]:
        with self.wait_condition:
            ret = []
            for value in self.values:
                if value not in self.in_flight:
                    ret.append(value)
            return ret

    def find(self, func: Callable[[T], bool]) -> Optional[T]:
        with self.wait_condition:
            for value in self.values:
                if func(value):
                    return value
            return None

    def remove(self, value: T) -> None:
        with self.wait_condition:
            if value in self.in_flight:
                del self.in_flight[value]
            try:
                self.values.remove(value)
            except ValueError:
                pass

    def remove_from_in_flight(self, value: T) -> None:
        with self.wait_condition:
            if value in self.in_flight:
                del self.in_flight[value]
                self.wait_condition.notify_all()

    def wait(self):
        with self.wait_condition:
            if self.stopped:
                return

            if len(self.values) > len(self.in_flight):
                return

            self.wait_condition.wait()

    def stop(self):
        with self.wait_condition:
            self.stopped = True
            self.wait_condition.notify_all()

    def sleep(self, timeout: float) -> None:
        with self.wait_condition:
            if self.stopped:
                return
            self.wait_condition.wait(timeout=timeout)

    def __enter__(self):
        self.wait_condition.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait_condition.release()



class BOHOrderPictureRequest(object):
    def __init__(
        self,
        pos_id,    # type: str
        order_id,  # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.order_id = order_id



from datetime import datetime

from ._TagEventType import TagEventType


class TagEvent(object):
    def __init__(self, tag, action, date, hallow=False):
        # type: (str, TagEventType, datetime, bool) -> None
        self.tag = tag
        self.action = action
        self.date = date
        self.hallow = hallow



from abc import ABCMeta, abstractmethod
from datetime import datetime

from production.model import ProductionOrder


class UndoCommand(object):
    __metaclass__ = ABCMeta

    def __init__(self, order_id, view):
        # type: (str, str) -> None

        self.order_id = order_id
        self.view = view

        self.enabled = True
        self.timestamp = datetime.now()

    @abstractmethod
    def undo(self, order):
        # type: (ProductionOrder) -> None

        raise NotImplementedError()

    def disable(self):
        self.enabled = False

    def enable(self):
        self.enabled = True

    def is_enabled(self):
        return self.enabled



# DER decoder
from pyasn1.codec.cer import decoder

tagMap = decoder.tagMap
typeMap = decoder.typeMap
class Decoder(decoder.Decoder):
    supportIndefLength = False

decode = Decoder(tagMap, typeMap)



import pyscripts

from boss_api import Boss
from mbcontextmessagehandler import MbContextMessageBus


def get_boss_service():
    # type: (...) -> Boss

    message_bus = MbContextMessageBus(mbcontext=pyscripts.mbcontext)
    return Boss(message_bus=message_bus)



class CustomProperty(object):

    def __init__(
        self,
        key=None,      # type: str
        value=None     # type: str
    ):
        # type: (...) -> None

        self.key = key
        self.value = value



import json

from msgbus import (
    FM_PARAM,
    MBEasyContext,
    TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES,
)
from typing import (
    Optional,
    Dict,
)


class OrderService(object):

    def __init__(
        self,           # type: OrderService
        mb_context,     # type: MBEasyContext
    ):
        # type: (...) -> None

        self.mb_context = mb_context

    def set_custom_property(
        self,               # type: OrderService
        pos_id,             # type: str
        key,                # type: str
        value,              # type: str
        order_id="",        # type: str
        blk_notify=False,   # type: Optional[bool]
    ):
        # type: (...) -> None

        # TODO: Change to order_api.OrderService. Did not make by update to MbContextMessageHandler==4.0.0 cause error

        request = {
            "customProperties": {key: value},
            "posId": int(pos_id),
            "orderId": order_id,
            "lineNumber": None,
            "itemId": None,
            "level": None,
            "partCode": None,
            "blkNotify": blk_notify,
        }

        self.mb_context.MB_EasySendMessage(
            dest_name="ORDERMGR",
            token=TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES,
            format=FM_PARAM,
            data=json.dumps(request).encode("utf-8"),
            timeout=100 * 100000,
        )

    def set_custom_properties(
        self,               # type: OrderService
        pos_id,             # type: str
        custom_properties,  # type: Dict
        order_id,           # type: str
        blk_notify=False,   # type: Optional[bool]
    ):
        # type: (...) -> None

        # TODO: Change to order_api.OrderService. Did not make by update to MbContextMessageHandler==4.0.0 cause error
        request = {
            "customProperties": self._parse_properties(custom_properties=custom_properties),
            "posId": int(pos_id),
            "orderId": order_id,
            "lineNumber": None,
            "itemId": None,
            "level": None,
            "partCode": None,
            "blkNotify": blk_notify,
        }
        self.mb_context.MB_EasySendMessage(
            dest_name="ORDERMGR",
            token=TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES,
            format=FM_PARAM,
            data=json.dumps(request).encode("utf-8"),
            timeout=100 * 100000,
        )

    @staticmethod
    def _parse_properties(
        custom_properties,  # type: Dict[str, str]
    ):
        # type: (...) -> Dict[str, str]

        properties = {}
        for key, value in custom_properties.items():
            try:
                if isinstance(value, unicode):  # noqa
                    value = value.encode("utf-8")
                fixed_value = str(value).decode("utf-8")
            except (Exception,):
                fixed_value = str(value)
            properties[key] = fixed_value

        return properties



from logging import Logger

from production.model import (
    ProductionOrder,
)
from production.util import parse_constants
from production.view.printerviewservices.reportgenerator import PickListReportGenerator
from production.view.printerviewservices.reportprinter import PickListPrinter
from typing import List
from ._PrinterView import PrinterView


class PicklistPrinterView(PrinterView):
    def __init__(
        self,
        name,                   # type: str
        production_manager,     # type: ProductionManager
        report_generator,       # type: PickListReportGenerator
        picklist_printer,       # type: PickListPrinter
        tags_on_print=None,     # type: List[str]
        logger=None,            # type: Logger
    ):
        # type: (...) -> None

        super(PicklistPrinterView, self).__init__(
            name=name,
            tags_on_print=tags_on_print,
            logger=logger,
        )

        self.production_manager = production_manager
        self.report_generator = report_generator
        self.picklist_printer = picklist_printer

    def get_printer_name(
        self,       # type: PicklistPrinterView
    ):
        # type: (...) -> str

        return self.picklist_printer.get_name()

    def generate_report(
        self,       # type: PicklistPrinterView
        order,      # type: ProductionOrder
    ):
        # type: (...) -> str

        is_reprint = self.production_manager.get_picklist_id(order=order) is not None
        report = self.report_generator.production_pick_list(
            production_order=order,
            is_reprint=is_reprint,
        )
        return report

    def print_report(
        self,       # type: PicklistPrinterView
        report,     # type: str
        order,      # type: ProductionOrder
    ):
        # type: (...) -> None

        is_reprint = self.production_manager.get_picklist_id(order=order) is not None
        parsed_report = parse_constants(
            printer_text=report,
            pos_id=int(order.originator[-2:]),
            is_reprint=is_reprint,
        )
        return self.picklist_printer.print_report(text=parsed_report)

    def save_report(
        self,               # type: PrinterView
        report,             # type: str
        order,              # type: ProductionOrder
        print_status,       # type: bool
    ):
        # type: (...) -> None

        self.production_manager.save_pick_list_order(
            order=order,
            receipt=report,
            status=print_status,
        )

    def clear_printer_cache(
        self,       # type: PicklistPrinterView
    ):
        # type: (...) -> None

        self.picklist_printer.clear_printer_cache()



# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors


from ... import exc
from ... import util
from ...sql import coercions
from ...sql import elements
from ...sql import operators
from ...sql import roles
from ...sql.base import _generative
from ...sql.base import Generative
from ...util.typing import Self


class match(Generative, elements.BinaryExpression):
    """Produce a ``MATCH (X, Y) AGAINST ('TEXT')`` clause.

    E.g.::

        from sqlalchemy import desc
        from sqlalchemy.dialects.mysql import match

        match_expr = match(
            users_table.c.firstname,
            users_table.c.lastname,
            against="Firstname Lastname",
        )

        stmt = (
            select(users_table)
            .where(match_expr.in_boolean_mode())
            .order_by(desc(match_expr))
        )

    Would produce SQL resembling::

        SELECT id, firstname, lastname
        FROM user
        WHERE MATCH(firstname, lastname) AGAINST (:param_1 IN BOOLEAN MODE)
        ORDER BY MATCH(firstname, lastname) AGAINST (:param_2) DESC

    The :func:`_mysql.match` function is a standalone version of the
    :meth:`_sql.ColumnElement.match` method available on all
    SQL expressions, as when :meth:`_expression.ColumnElement.match` is
    used, but allows to pass multiple columns

    :param cols: column expressions to match against

    :param against: expression to be compared towards

    :param in_boolean_mode: boolean, set "boolean mode" to true

    :param in_natural_language_mode: boolean , set "natural language" to true

    :param with_query_expansion: boolean, set "query expansion" to true

    .. versionadded:: 1.4.19

    .. seealso::

        :meth:`_expression.ColumnElement.match`

    """

    __visit_name__ = "mysql_match"

    inherit_cache = True

    def __init__(self, *cols, **kw):
        if not cols:
            raise exc.ArgumentError("columns are required")

        against = kw.pop("against", None)

        if against is None:
            raise exc.ArgumentError("against is required")
        against = coercions.expect(
            roles.ExpressionElementRole,
            against,
        )

        left = elements.BooleanClauseList._construct_raw(
            operators.comma_op,
            clauses=cols,
        )
        left.group = False

        flags = util.immutabledict(
            {
                "mysql_boolean_mode": kw.pop("in_boolean_mode", False),
                "mysql_natural_language": kw.pop(
                    "in_natural_language_mode", False
                ),
                "mysql_query_expansion": kw.pop("with_query_expansion", False),
            }
        )

        if kw:
            raise exc.ArgumentError("unknown arguments: %s" % (", ".join(kw)))

        super().__init__(left, against, operators.match_op, modifiers=flags)

    @_generative
    def in_boolean_mode(self) -> Self:
        """Apply the "IN BOOLEAN MODE" modifier to the MATCH expression.

        :return: a new :class:`_mysql.match` instance with modifications
         applied.
        """

        self.modifiers = self.modifiers.union({"mysql_boolean_mode": True})
        return self

    @_generative
    def in_natural_language_mode(self) -> Self:
        """Apply the "IN NATURAL LANGUAGE MODE" modifier to the MATCH
        expression.

        :return: a new :class:`_mysql.match` instance with modifications
         applied.
        """

        self.modifiers = self.modifiers.union({"mysql_natural_language": True})
        return self

    @_generative
    def with_query_expansion(self) -> Self:
        """Apply the "WITH QUERY EXPANSION" modifier to the MATCH expression.

        :return: a new :class:`_mysql.match` instance with modifications
         applied.
        """

        self.modifiers = self.modifiers.union({"mysql_query_expansion": True})
        return self



from logging import Logger

from production.model import Item
from typing import (
    Dict,
    Optional,
    Union,
    List,
)
from ._OrderChangerProductionBox import OrderChangerProductionBox
from ._ProductionBox import ProductionBox


class CollapseSameItemsBox(OrderChangerProductionBox):
    def __init__(self, name, sons, include_seats, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], bool, Logger) -> None

        super(CollapseSameItemsBox, self).__init__(name, sons, logger)

        self.include_seats = include_seats

    def change_order(self, order):
        new_items = []
        collapsed_items = {}  # type: Dict[str, Item]
        for item in order.items:
            key = self.build_item_key(item)
            if key not in collapsed_items:
                collapsed_items[key] = item
                if not hasattr(item, 'joined_items'):
                    item.joined_items = []

                new_items.append(item)
            else:
                self._update_collapsed_sons_quantities(
                    collapsed_item=collapsed_items[key],
                    item=item,
                )

        order.items = new_items
        return order

    def _update_collapsed_sons_quantities(
        self,               # type: CollapseSameItemsBox
        collapsed_item,     # type: Item
        item,               # type: Item
    ):
        # type: (...) -> None

        collapsed_item.qty += item.qty
        collapsed_item.multiplied_qty += item.multiplied_qty
        collapsed_item.joined_items.append(item)

        for collapsed_son in collapsed_item.items:
            for son in item.items:
                if collapsed_son.part_code == son.part_code and collapsed_son.level == son.level:
                    self._update_collapsed_sons_quantities(
                        collapsed_item=collapsed_son,
                        item=son,
                    )

    def build_item_key(self, item, key="", is_parent_product=False):
        # type: (Item, str, bool) -> str

        key += str(item.part_code)
        if self.include_seats:
            key += item.seat

        if len(item.comments) > 0:
            for comment in item.comments.values():
                key += comment.text

        if is_parent_product:
            key += str(item.qty)

        for tag in item.get_tags():
            key += tag

        for son in item.items:
            is_parent_product = item.item_type.upper() == "PRODUCT"
            key += self.build_item_key(son, key, is_parent_product=is_parent_product)

        return key



import json
import logging

from helper import remove_accents
from old_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection
from typing import Dict

logger = logging.getLogger("FiscalWrapper")


class ProductRepository(BaseRepository):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        super(ProductRepository, self).__init__(mb_context)

    def get_tender_types(self):
        # type: () -> Dict

        def inner_func(
            conn,       # type: (Connection)
        ):
            # type: (...) -> Dict

            tender_types_dict = {}
            try:
                query = """SELECT TenderId, TenderDescr, CustomAttr FROM TenderType """

                for row in conn.select(query):
                    tender_id = row.get_entry("TenderId")
                    description = remove_accents(row.get_entry("TenderDescr"))
                    custom_attr = json.loads(row.get_entry("CustomAttr") or {})

                    tender_types_dict[description.upper()] = {
                        "tenderId": tender_id,
                        "description": description,
                    }

                    self._set_dict_tender_type(
                        custom_attr=custom_attr,
                        custom_attr_key="CNPJ",
                        tender_types_dict=tender_types_dict,
                        tender_types_dict_key="cnpj",
                        description=description,
                    )
                    self._set_dict_tender_type(
                        custom_attr=custom_attr,
                        custom_attr_key="clientId",
                        tender_types_dict=tender_types_dict,
                        tender_types_dict_key="clientId",
                        description=description,
                    )
                    self._set_dict_tender_type(
                        custom_attr=custom_attr,
                        custom_attr_key="tBand",
                        tender_types_dict=tender_types_dict,
                        tender_types_dict_key="tBand",
                        description=description,
                    )
                    self._set_dict_tender_type(
                        custom_attr=custom_attr,
                        custom_attr_key="tPag",
                        tender_types_dict=tender_types_dict,
                        tender_types_dict_key="tPag",
                        description=description,
                    )
                    self._set_dict_tender_type(
                        custom_attr=custom_attr,
                        custom_attr_key="cAdmC",
                        tender_types_dict=tender_types_dict,
                        tender_types_dict_key="cAdmC",
                        description=description,
                    )

                return tender_types_dict
            except (Exception,):
                logger.exception('Error to get tenders types')
                return tender_types_dict

        return self.execute_with_connection(inner_func)

    @ staticmethod
    def _set_dict_tender_type(
        custom_attr,                # type: Dict
        custom_attr_key,            # type: str
        tender_types_dict,          # type: Dict
        tender_types_dict_key,      # type: str
        description,                # type: str
    ):
        # type: (...) -> None

        custom_attr_value = custom_attr.get(custom_attr_key, None)

        if custom_attr_value:
            tender_types_dict[description.upper()][tender_types_dict_key] = str(custom_attr_value)
